﻿''' <summary>
''' A window to represent a client that sends text messages to and receives text messages from a server.
''' </summary>
Public Class ClientWindow

#Region " Fields "

    ''' <summary>
    ''' The name of the server to connect to.
    ''' </summary>
    ''' <remarks>
    ''' In this case use the local machine.
    ''' </remarks>
    Private ReadOnly host As String = Environment.MachineName
    ''' <summary>
    ''' The port number on the server to connect to.
    ''' </summary>
    Private ReadOnly port As Integer = 12345
    ''' <summary>
    ''' The client object that will send the messages to the server.
    ''' </summary>
    Private WithEvents client As New MessageClient(host, port)

#End Region 'Fields

#Region " Event Handlers "

    Private Sub ClientWindow_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        'Destroy the client when the form closes.
        Me.client.Dispose()
    End Sub


    Private Sub connectButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles connectButton.Click
        'Connecting can be done only once.
        Me.connectButton.Enabled = False

        'Connect to the server.
        Me.client.Connect()
    End Sub

    Private Sub sendButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles sendButton.Click
        Dim message = Me.messageTextBox.Text

        'Send the message to the server.
        Me.client.Send(message)

        'Update the display.
        Me.UpdateLog(String.Format("Message (=>{0}:{1}): {2}", Me.host, Me.port, message))
        Me.messageTextBox.SelectAll()
        Me.messageTextBox.Select()
    End Sub


    Private Sub client_ConnectionFailed(ByVal sender As Object, ByVal e As Net.ConnectionEventArgs) Handles client.ConnectionFailed
        Me.PromptToReconnect("The specified server could not be found.  Would you like to try again?", _
                             "Connection Failed", _
                             MessageBoxIcon.Error)
    End Sub

    Private Sub client_ConnectionAccepted(ByVal sender As Object, ByVal e As Net.ConnectionEventArgs) Handles client.ConnectionAccepted
        Dim host = e.Host.ToString()

        'Update the display.
        Me.Text = String.Format("{0}=>{1}", Me.client.LocalPort, host)
        Me.UpdateLog("Info: Connection accepted to " & host)

        'The connection is open so allow messages to be sent.
        Me.sendButton.Enabled = True
    End Sub

    Private Sub client_MessageReceived(ByVal sender As Object, ByVal e As Net.MessageReceivedEventArgs) Handles client.MessageReceived
        'Update the display with the message received from the server.
        Me.UpdateLog(String.Format("Message ({0}=>): {1}", e.Host, e.Message))
    End Sub

    Private Sub client_ConnectionClosed(ByVal sender As Object, ByVal e As Net.ConnectionEventArgs) Handles client.ConnectionClosed
        Me.PromptToReconnect("The connection was closed by the server.  Would you like to reconnect?", _
                             "Connection Closed", _
                             MessageBoxIcon.Warning)
    End Sub

#End Region 'Event Handlers

#Region " Methods "

    ''' <summary>
    ''' Prompts the user to try again to connect to the server.
    ''' </summary>
    ''' <param name="message">
    ''' The error message to display.
    ''' </param>
    ''' <param name="caption">
    ''' The title to display on the message box.
    ''' </param>
    ''' <param name="icon">
    ''' The icon to display on the message box.
    ''' </param>
    ''' <remarks>
    ''' If the user cancels the action the client window will be closed.
    ''' </remarks>
    Private Sub PromptToReconnect(ByVal message As String, ByVal caption As String, ByVal icon As MessageBoxIcon)
        Me.sendButton.Enabled = False
        Me.connectButton.Enabled = True

        Select Case MessageBox.Show(message, _
                                    caption, _
                                    MessageBoxButtons.YesNoCancel, _
                                    icon)
            Case Windows.Forms.DialogResult.Yes
                'Try again to connect.
                Me.connectButton.PerformClick()
            Case Windows.Forms.DialogResult.No
                'Do nothing.
            Case Windows.Forms.DialogResult.Cancel
                Me.Close()
        End Select
    End Sub

    ''' <summary>
    ''' Appends text to the log field, followed by a line break.
    ''' </summary>
    ''' <param name="text">
    ''' The text to append.
    ''' </param>
    Private Sub UpdateLog(ByVal text As String)
        Me.logTextBox.AppendText(text & ControlChars.NewLine)
    End Sub

#End Region 'Methods

End Class