﻿''' <summary>
''' A window to represent a server that sends text messages to and receives text messages from multiple clients.
''' </summary>
Public Class MainWindow

#Region " Fields "

    ''' <summary>
    ''' The server object that will receive the messages from the clients.
    ''' </summary>
    ''' <remarks>
    ''' Use the current form as the synchronising object so that events are raised in the UI thread.
    ''' </remarks>
    Private WithEvents server As New MessageServer(12345)
    ''' <summary>
    ''' Contains client information keyed on the string representation displayed in the client drop-down list.
    ''' </summary>
    Private ReadOnly hosts As New List(Of HostInfo)

#End Region 'Fields

#Region " Event Handlers "

    Private Sub MainWindow_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        'Destroy the server when the form closes.
        Me.server.Dispose()
    End Sub


    Private Sub sendButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles sendButton.Click
        Dim message = Me.messageTextBox.Text
        Dim host = DirectCast(Me.hostsComboBox.SelectedItem, HostInfo)

        'Send the message to the selected client.
        Me.server.Send(host, message)

        'Update the display.
        Me.UpdateLog(String.Format("Message (=>{0}): {1}", host, message))
        Me.messageTextBox.SelectAll()
        Me.messageTextBox.Select()
    End Sub

    Private Sub sendAllButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles sendAllButton.Click
        Dim message = Me.messageTextBox.Text

        'Send the message to all clients.
        Me.server.Send(message)

        'Update the display.
        Me.UpdateLog("Message (=>All): " & message)
        Me.messageTextBox.SelectAll()
        Me.messageTextBox.Select()
    End Sub


    Private Sub server_ConnectionAccepted(ByVal sender As Object, ByVal e As Net.ConnectionEventArgs) Handles server.ConnectionAccepted
        Dim host = e.Host

        'Update the display.
        Me.UpdateLog("Info: Connection accepted from " & host.ToString())
        Me.hostsComboBox.Items.Add(host)
        Me.UpdateSendButtonsState()

        'Select the new client if it is the first.
        If Me.hostsComboBox.SelectedItem Is Nothing Then
            Me.hostsComboBox.SelectedIndex = 0
        End If

        'Remember this client.
        Me.hosts.Add(host)
    End Sub

    Private Sub server_MessageReceived(ByVal sender As Object, ByVal e As Net.MessageReceivedEventArgs) Handles server.MessageReceived
        'Update the display with the message received from the client.
        Me.UpdateLog(String.Format("Message ({0}=>): {1}", e.Host, e.Message))
    End Sub

    Private Sub server_ConnectionClosed(ByVal sender As Object, ByVal e As Net.ConnectionEventArgs) Handles server.ConnectionClosed
        Dim host = e.Host
        Dim selectedIndex = Me.hostsComboBox.SelectedIndex
        Dim updateHostSelection = (host.Equals(Me.hostsComboBox.SelectedItem))

        'Remove the disconnected host.
        Me.hostsComboBox.Items.Remove(host)

        'Update the display.
        Me.UpdateLog(String.Format("Info: Connection from {0} closed", host))
        Me.hosts.Remove(host)
        Me.UpdateSendButtonsState()

        'Select a new host.
        If updateHostSelection Then
            Select Case Me.hostsComboBox.Items.Count
                Case 0
                    'No hosts to select.
                Case selectedIndex
                    'The previous selection was the last in the list so select the new last item.
                    Me.hostsComboBox.SelectedIndex = selectedIndex - 1
                Case Else
                    'The previous selection was not the last in the list so select the host at the same index.
                    Me.hostsComboBox.SelectedIndex = selectedIndex
            End Select
        End If
    End Sub

#End Region 'Event Handlers

#Region " Methods "

    ''' <summary>
    ''' Appends text to the log field, followed by a line break.
    ''' </summary>
    ''' <param name="text">
    ''' The text to append.
    ''' </param>
    Private Sub UpdateLog(ByVal text As String)
        Me.logTextBox.AppendText(text & ControlChars.NewLine)
    End Sub

    ''' <summary>
    ''' Enables the <i>Send</i> and <i>Send All</i> buttons if and only if there is at least one client connected.
    ''' </summary>
    Private Sub UpdateSendButtonsState()
        Dim enabled = (Me.hostsComboBox.Items.Count > 0)

        Me.sendButton.Enabled = enabled
        Me.sendAllButton.Enabled = enabled
    End Sub

#End Region 'Methods

End Class
