﻿''' <summary>
''' Initiates a conection with a server and sends ands receives text messages over that connection.
''' </summary>
Public Class MessageClient
    Inherits MessageClientServerBase

#Region " Fields "

    ''' <summary>
    ''' The client used to communicate with the server
    ''' </summary>
    Private ReadOnly client As New TcpClient
    ''' <summary>
    ''' Indicates whether or not the current instance has been disposed.
    ''' </summary>
    Private isDisposed As Boolean = False
    ''' <summary>
    ''' The port number of the local end of the connection.
    ''' </summary>
    Private _localPort As Integer
    ''' <summary>
    ''' The details of the server to connect to.
    ''' </summary>
    Private server As HostInfo
    ''' <summary>
    ''' The stream over which messages are sent and received.
    ''' </summary>
    Private stream As NetworkStream

#End Region 'Fields

#Region " Properties "

    ''' <summary>
    ''' Gets the port number of the local end of the connection.
    ''' </summary>
    ''' <value>
    ''' An <b>Int32</b> containing a port number.
    ''' </value>
    Public ReadOnly Property LocalPort() As Integer
        Get
            Return Me._localPort
        End Get
    End Property

#End Region 'Properties

#Region " Constructors "

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClient" /> class with the default buffer size and character encoding.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the server to connect to.
    ''' </param>
    ''' <param name="remotePort">
    ''' The port number to connect to.
    ''' </param>
    Public Sub New(ByVal hostName As String, ByVal remotePort As Integer)
        Me.Initialise(hostName, remotePort)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClient" /> class with the default character encoding.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the server to connect to.
    ''' </param>
    ''' <param name="remotePort">
    ''' The port number to connect to.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    Public Sub New(ByVal hostName As String, ByVal remotePort As Integer, ByVal bufferSize As Integer)
        MyBase.New(bufferSize)
        Me.Initialise(hostName, remotePort)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClient" /> class with the default buffer size.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the server to connect to.
    ''' </param>
    ''' <param name="remotePort">
    ''' The port number to connect to.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal hostName As String, ByVal remotePort As Integer, ByVal encoding As Encoding)
        MyBase.New(encoding)
        Me.Initialise(hostName, remotePort)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClient" /> class.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the server to connect to.
    ''' </param>
    ''' <param name="remotePort">
    ''' The port number to connect to.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal hostName As String, ByVal remotePort As Integer, ByVal bufferSize As Integer, ByVal encoding As Encoding)
        MyBase.New(bufferSize, encoding)
        Me.Initialise(hostName, remotePort)
    End Sub

#End Region 'Constructors

#Region " Events "

    ''' <summary>
    ''' Occurs when a connection attempt fails.
    ''' </summary>
    Public Event ConnectionFailed As ConnectionEventHandler

#End Region 'Events

#Region " Methods "

#Region " Public Methods "

    ''' <summary>
    ''' Connects to the server.
    ''' </summary>
    Public Sub Connect()
        'Connect asynchronously to the server.
        Me.client.BeginConnect(Me.server.HostName, Me.server.Port, AddressOf Connect, Nothing)
    End Sub

    ''' <summary>
    ''' Sends a message to the server.
    ''' </summary>
    ''' <param name="message">
    ''' A <b>String</b> containing the message to send.
    ''' </param>
    Public Sub Send(ByVal message As String)
        'Convert the text message to binary data.
        Dim buffer As Byte() = Me.Encoding.GetBytes(message)

        'Write the data asynchronously.
        Me.stream.BeginWrite(buffer, 0, buffer.Length, AddressOf Write, Nothing)
    End Sub

#End Region 'Public Methods

#Region " Protected Methods "

    ''' <summary>
    ''' Releases all resources used by the object.
    ''' </summary>
    ''' <param name="disposing">
    ''' <b>true</b> to release both managed and unmanaged resources; <b>false</b> to release only unmanaged resources.
    ''' </param>
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If Not Me.isDisposed Then
            If disposing Then
                'Close the connection and the underlying stream.
                If Me.stream IsNot Nothing Then
                    Me.stream.Close()
                End If

                Me.client.Close()
            End If

            Me.stream = Nothing
        End If

        MyBase.Dispose(disposing)
        Me.isDisposed = True
    End Sub

    ''' <summary>
    ''' Raises the <see cref="ConnectionFailed" /> event.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    ''' <remarks>
    ''' The event will be raised on the thread on which the current instance was created.
    ''' </remarks>
    Protected Overridable Sub OnConnectionFailed(ByVal e As ConnectionEventArgs)
        Me.SynchronisingContext.Post(AddressOf RaiseConnectionFailed, e)
    End Sub

#End Region 'Protected Methods

#Region " Private Methods "

    ''' <summary>
    ''' Completes a connection to the server.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the connection operation.
    ''' </param>
    Private Sub Connect(ByVal ar As IAsyncResult)
        Try
            'Complete the asynchronous connection.
            Me.client.EndConnect(ar)

            'Get the port that was assigned to the local end point.
            Me._localPort = DirectCast(Me.client.Client.LocalEndPoint, IPEndPoint).Port

            Me.stream = Me.client.GetStream()

            Dim buffer(Me.BufferSize - 1) As Byte

            'Listen asynchronously for an incoming message.
            Me.stream.BeginRead(buffer, 0, Me.BufferSize, AddressOf Read, buffer)

            'Notify any listeners that the connection was successful.
            Me.OnConnectionAccepted(New ConnectionEventArgs(Me.server))
        Catch ex As SocketException
            'The specified server was not found.
            Me.OnConnectionFailed(New ConnectionEventArgs(Me.server))
        End Try
    End Sub

    ''' <summary>
    ''' Creates a <see cref="HostInfo" /> to represent the server.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the server.
    ''' </param>
    ''' <param name="remotePort">
    ''' The port number of the server.
    ''' </param>
    Private Sub Initialise(ByVal hostName As String, ByVal remotePort As Integer)
        Me.server = New HostInfo(hostName, remotePort)
    End Sub

    ''' <summary>
    ''' Raises the <see cref="ConnectionFailed" /> event on the current thread.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    Private Sub RaiseConnectionFailed(ByVal e As Object)
        RaiseEvent ConnectionFailed(Me, DirectCast(e, ConnectionEventArgs))
    End Sub

    ''' <summary>
    ''' Receives an incoming message.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the read operation.
    ''' </param>
    Private Sub Read(ByVal ar As IAsyncResult)
        Try
            'The stream will be Nothing if the client has been disposed.
            If Me.stream IsNot Nothing Then
                Dim buffer = DirectCast(ar.AsyncState, Byte())

                'Complete the asynchronous read and get the first block of data.
                Dim byteCount = Me.stream.EndRead(ar)

                If byteCount = 0 Then
                    'If there is no data when an asynchronous read completes it is because the server closed the connection.
                    Me.OnConnectionClosed(New ConnectionEventArgs(Me.server))
                Else
                    'Start building the message.
                    Dim message As New StringBuilder(Me.Encoding.GetString(buffer, 0, byteCount))

                    'As long as there is more data...
                    While Me.stream.DataAvailable
                        '...read another block of data.
                        byteCount = Me.stream.Read(buffer, 0, Me.BufferSize)

                        'Build the message block by block.
                        message.Append(Me.Encoding.GetString(buffer, 0, byteCount))
                    End While

                    'Listen asynchronously for another incoming message.
                    Me.stream.BeginRead(buffer, 0, Me.BufferSize, AddressOf Read, buffer)

                    'Notify any listeners that a message was received.
                    Me.OnMessageReceived(New MessageReceivedEventArgs(Me.server, message.ToString()))
                End If
            End If
        Catch ex As IOException
            'The callback specified when BeginRead was called may get invoked one last time when the TcpClient is disposed.
            'This exception is thrown when EndRead is called on a disposed client stream.
        End Try
    End Sub

    ''' <summary>
    ''' Completes writing a message to the server.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the write operation.
    ''' </param>
    ''' <remarks></remarks>
    Private Sub Write(ByVal ar As IAsyncResult)
        'Complete the asynchronous write.
        Me.stream.EndWrite(ar)
    End Sub

#End Region 'Private Methods

#End Region 'Methods

End Class
