﻿''' <summary>
''' Acts as a base class for both server and client types who can send text messages to each other.
''' </summary>
Public MustInherit Class MessageClientServerBase
    Implements IDisposable

#Region " Fields "

    ''' <summary>
    ''' The block size in which to read incoming messages.
    ''' </summary>
    Private ReadOnly _bufferSize As Integer = 1024
    ''' <summary>
    ''' The encoding to use when converting between binary data to text.
    ''' </summary>
    Private ReadOnly _encoding As Encoding = System.Text.Encoding.Unicode
    ''' <summary>
    ''' Indicates whether or not the current instance has been disposed.
    ''' </summary>
    Private isDisposed As Boolean = False
    ''' <summary>
    ''' The object used to marshal event handler calls.
    ''' </summary>
    ''' <remarks>
    ''' All events will be raised on the thread on which the current instance was created.
    ''' </remarks>
    Private _synchronisingContext As SynchronizationContext = SynchronizationContext.Current

#End Region 'Fields

#Region " Properties "

    ''' <summary>
    ''' Gets the size of the blocks in which incoming messages are read.
    ''' </summary>
    ''' <value>
    ''' An <b>Int32</b> containing the read block size.  The default is 1024 (1 KB).
    ''' </value>
    Protected ReadOnly Property BufferSize() As Integer
        Get
            Return Me._bufferSize
        End Get
    End Property

    ''' <summary>
    ''' Gets the encoding used to convert text messages to binary data for transmission.
    ''' </summary>
    ''' <value>
    ''' The <see cref="System.Text.Encoding">Encoding</see> specified when the instance was created, or <see cref="Encoding.Unicode">Unicode</see> if an encoding was not specified.
    ''' </value>
    Protected ReadOnly Property Encoding() As Encoding
        Get
            Return Me._encoding
        End Get
    End Property

    ''' <summary>
    ''' Gets the encoding used to convert text messages to binary data for transmission.
    ''' </summary>
    ''' <value>
    ''' The <see cref="System.Text.Encoding">Encoding</see> specified when the instance was created, or <see cref="Encoding.Unicode">Unicode</see> if an encoding was not specified.
    ''' </value>
    Protected ReadOnly Property SynchronisingContext() As SynchronizationContext
        Get
            Return Me._synchronisingContext
        End Get
    End Property

#End Region 'Properties

#Region " Constructors "

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClientServerBase" /> class with the default buffer size and character encoding.
    ''' </summary>
    Public Sub New()
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClientServerBase" /> class with the default character encoding.
    ''' </summary>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    Public Sub New(ByVal bufferSize As Integer)
        Me._bufferSize = bufferSize
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClientServerBase" /> class with the default buffer size.
    ''' </summary>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal encoding As Encoding)
        Me._encoding = encoding
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageClientServerBase" /> class.
    ''' </summary>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal bufferSize As Integer, ByVal encoding As Encoding)
        Me._bufferSize = bufferSize
        Me._encoding = encoding
    End Sub

#End Region 'Constructors

#Region " Events "

    ''' <summary>
    ''' Occurs when a connection attempt is accepted.
    ''' </summary>
    Public Event ConnectionAccepted As ConnectionEventHandler
    ''' <summary>
    ''' Occurs when a connection is closed.
    ''' </summary>
    Public Event ConnectionClosed As ConnectionEventHandler
    ''' <summary>
    ''' Occurs when a message is received.
    ''' </summary>
    Public Event MessageReceived As MessageReceivedEventHandler

#End Region 'Events

#Region " Methods "

#Region " Public Methods "

    ''' <summary>
    ''' Releases all resources used by the object.
    ''' </summary>
    ''' <remarks>
    ''' This method should be called when the object is no longer needed.
    ''' </remarks>
    Public Sub Dispose() Implements IDisposable.Dispose
        ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

#End Region 'Public Methods

#Region " Protected Methods "

    ''' <summary>
    ''' Releases all resources used by the object.
    ''' </summary>
    ''' <param name="disposing">
    ''' <b>true</b> to release both managed and unmanaged resources; <b>false</b> to release only unmanaged resources.
    ''' </param>
    Protected Overridable Sub Dispose(ByVal disposing As Boolean)
        If Not Me.isDisposed Then
            If disposing Then
            End If

            Me._synchronisingContext = Nothing
        End If

        Me.isDisposed = True
    End Sub

    ''' <summary>
    ''' Raises the <see cref="ConnectionAccepted" /> event.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    ''' <remarks>
    ''' The event will be raised on the thread on which the current instance was created.
    ''' </remarks>
    Protected Overridable Sub OnConnectionAccepted(ByVal e As ConnectionEventArgs)
        Me._synchronisingContext.Post(AddressOf RaiseConnectionAccepted, e)
    End Sub

    ''' <summary>
    ''' Raises the <see cref="ConnectionClosed" /> event.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    ''' <remarks>
    ''' The event will be raised on the thread on which the current instance was created.
    ''' </remarks>
    Protected Overridable Sub OnConnectionClosed(ByVal e As ConnectionEventArgs)
        Me._synchronisingContext.Post(AddressOf RaiseConnectionClosed, e)
    End Sub

    ''' <summary>
    ''' Raises the <see cref="MessageReceived" /> event.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    ''' <remarks>
    ''' The event will be raised on the thread on which the current instance was created.
    ''' </remarks>
    Protected Overridable Sub OnMessageReceived(ByVal e As MessageReceivedEventArgs)
        Me._synchronisingContext.Post(AddressOf RaiseMessageReceived, e)
    End Sub

#End Region 'Protected Methods

#Region " Private Methods "


    ''' <summary>
    ''' Raises the <see cref="ConnectionAccepted" /> event on the current thread.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    Private Sub RaiseConnectionAccepted(ByVal e As Object)
        RaiseEvent ConnectionAccepted(Me, DirectCast(e, ConnectionEventArgs))
    End Sub

    ''' <summary>
    ''' Raises the <see cref="ConnectionClosed" /> event on the current thread.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    Private Sub RaiseConnectionClosed(ByVal e As Object)
        RaiseEvent ConnectionClosed(Me, DirectCast(e, ConnectionEventArgs))
    End Sub

    ''' <summary>
    ''' Raises the <see cref="MessageReceived" /> event on the current thread.
    ''' </summary>
    ''' <param name="e">
    ''' Contains the data for the event.
    ''' </param>
    Private Sub RaiseMessageReceived(ByVal e As Object)
        RaiseEvent MessageReceived(Me, DirectCast(e, MessageReceivedEventArgs))
    End Sub

#End Region 'Private Methods

#End Region 'Methods

End Class
