﻿''' <summary>
''' Receives connection requests from clients and receives and sends text messages over those connections.
''' </summary>
Public Class MessageServer
    Inherits MessageClientServerBase

#Region " Types "

    ''' <summary>
    ''' Contains application-specific state data for an asynchronous read.
    ''' </summary>
    Private Class ReadAsyncState
        ''' <summary>
        ''' The client receiving the data.
        ''' </summary>
        Public Client As TcpClient
        ''' <summary>
        ''' A <b>Byte</b> array the data is read into.
        ''' </summary>
        Public Buffer As Byte()
    End Class

#End Region 'Types

#Region " Fields "

    ''' <summary>
    ''' The connected clients and the corresponding remote host information.
    ''' </summary>
    Private ReadOnly clients As New Dictionary(Of TcpClient, HostInfo)
    ''' <summary>
    ''' Indicates whether or not the current instance has been disposed.
    ''' </summary>
    Private isDisposed As Boolean = False
    ''' <summary>
    ''' The port on which the server is listening for connections.
    ''' </summary>
    Private _port As Integer
    ''' <summary>
    ''' The server listening for connections.
    ''' </summary>
    ''' <remarks></remarks>
    Private server As TcpListener

#End Region 'Fields

#Region " Properties "

    ''' <summary>
    ''' Gets the port number on which the server is listening.
    ''' </summary>
    ''' <value>
    ''' An <b>Int32</b> containing a port number.
    ''' </value>
    Public ReadOnly Property Port() As Integer
        Get
            Return Me._port
        End Get
    End Property

    ''' <summary>
    ''' Gets a list of remote hosts connected to the server.
    ''' </summary>
    ''' <value>
    ''' An array of <see cref="HostInfo" /> objects containing the names and port numbers of the remote hosts.
    ''' </value>
    ''' <remarks>
    ''' The value of this property is generated ad hoc so the property value must be retrieved each and every time current data is required.
    ''' </remarks>
    Public ReadOnly Property Hosts() As HostInfo()
        Get
            Return Me.clients.Values.ToArray()
        End Get
    End Property

#End Region 'Properties

#Region " Constructors "

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding listening on any local IP address and a random port.
    ''' </summary>
    Public Sub New()
        Me.Initialise(IPAddress.Any, 0)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding listening on a random port.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address or host name on which to listen.
    ''' </param>
    Public Sub New(ByVal address As String)
        Me.Initialise(Me.ParseAddress(address), 0)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding listening on a random port.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address on which to listen.
    ''' </param>
    Public Sub New(ByVal address As IPAddress)
        Me.Initialise(address, 0)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding listening on any local IP address.
    ''' </summary>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    Public Sub New(ByVal port As Integer)
        Me.Initialise(IPAddress.Any, port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address or host name on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    Public Sub New(ByVal address As String, ByVal port As Integer)
        Me.Initialise(Me.ParseAddress(address), port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size and character encoding.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    Public Sub New(ByVal address As IPAddress, ByVal port As Integer)
        Me.Initialise(address, port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default character encoding.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address or host name on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    Public Sub New(ByVal address As String, _
                   ByVal port As Integer, _
                   ByVal bufferSize As Integer)
        MyBase.New(bufferSize)
        Me.Initialise(Me.ParseAddress(address), port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address or host name on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal address As String, _
                   ByVal port As Integer, _
                   ByVal encoding As Encoding)
        MyBase.New(encoding)
        Me.Initialise(Me.ParseAddress(address), port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address or host name on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal address As String, _
                   ByVal port As Integer, _
                   ByVal bufferSize As Integer, _
                   ByVal encoding As Encoding)
        MyBase.New(bufferSize, encoding)
        Me.Initialise(Me.ParseAddress(address), port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default character encoding.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    Public Sub New(ByVal address As IPAddress, _
                   ByVal port As Integer, _
                   ByVal bufferSize As Integer)
        MyBase.New(bufferSize)
        Me.Initialise(address, port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class with the default buffer size.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal address As IPAddress, _
                   ByVal port As Integer, _
                   ByVal encoding As Encoding)
        MyBase.New(encoding)
        Me.Initialise(address, port)
    End Sub

    ''' <summary>
    ''' Creates a new instance of the <see cref="MessageServer" /> class.
    ''' </summary>
    ''' <param name="address">
    ''' The IP address on which to listen.
    ''' </param>
    ''' <param name="port">
    ''' The port on which to listen for incoming connections.
    ''' </param>
    ''' <param name="bufferSize">
    ''' The block size in which to read incoming messages.
    ''' </param>
    ''' <param name="encoding">
    ''' The encoding to use when converting between binary data to text.
    ''' </param>
    Public Sub New(ByVal address As IPAddress, _
                   ByVal port As Integer, _
                   ByVal bufferSize As Integer, _
                   ByVal encoding As Encoding)
        MyBase.New(bufferSize, encoding)
        Me.Initialise(address, port)
    End Sub

#End Region 'Constructors

#Region " Methods "

#Region " Public Methods "

    ''' <summary>
    ''' Sends a message to the specified host.
    ''' </summary>
    ''' <param name="hostName">
    ''' The name or address of the host to send the message to.
    ''' </param>
    ''' <param name="port">
    ''' The port number of the host to send the message to.
    ''' </param>
    ''' <param name="message">
    ''' The message to send.
    ''' </param>
    Public Sub Send(ByVal hostName As String, ByVal port As Integer, ByVal message As String)
        Dim client = (From c In Me.clients.Keys _
                      Let h = Me.clients(c) _
                      Where h.HostName = hostName AndAlso h.Port = port _
                      Select c).First()

        Me.Send(client, message)
    End Sub

    ''' <summary>
    ''' Sends a message to the specified host.
    ''' </summary>
    ''' <param name="host">
    ''' The remote host to send the message to.
    ''' </param>
    ''' <param name="message">
    ''' The message to send.
    ''' </param>
    Public Sub Send(ByVal host As HostInfo, ByVal message As String)
        Dim client = (From c In Me.clients.Keys _
                      Where Me.clients(c).Equals(host) _
                      Select c).First()

        Me.Send(client, message)
    End Sub

    ''' <summary>
    ''' Sends a message to all connected hosts.
    ''' </summary>
    ''' <param name="message">
    ''' The message to send.
    ''' </param>
    Public Sub Send(ByVal message As String)
        For Each client In Me.clients.Keys
            Me.Send(client, message)
        Next
    End Sub

#End Region 'Public Methods

#Region " Protected Methods "

    ''' <summary>
    ''' Releases all resources used by the object.
    ''' </summary>
    ''' <param name="disposing">
    ''' <b>true</b> to release both managed and unmanaged resources; <b>false</b> to release only unmanaged resources.
    ''' </param>
    Protected Overrides Sub Dispose(ByVal disposing As Boolean)
        If Not Me.isDisposed Then
            If disposing Then
                'Close all remaining connections
                For Each client In Me.clients.Keys
                    client.GetStream().Close()
                    client.Close()
                Next

                'Stop listening for connection requests.
                Me.server.Stop()
            End If

            Me.clients.Clear()
        End If

        MyBase.Dispose(disposing)
        Me.isDisposed = True
    End Sub

#End Region 'Protected Methods

#Region " Private Methods "

    ''' <summary>
    ''' Accepts an incoming connection request.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the connection operation.
    ''' </param>
    Private Sub AcceptTcpClient(ByVal ar As IAsyncResult)
        Try
            Dim client = Me.server.EndAcceptTcpClient(ar)
            Dim host As HostInfo

            With DirectCast(client.Client.RemoteEndPoint, IPEndPoint)
                host = New HostInfo(.Address.ToString, .Port)
            End With

            'Remember the client and its host information.
            Me.clients.Add(client, host)

            'Listen asynchronously for more connections.
            Me.server.BeginAcceptTcpClient(AddressOf AcceptTcpClient, Nothing)

            Dim stream = client.GetStream()
            Dim buffer(Me.BufferSize - 1) As Byte

            'Listen asynchronously for incoming messages from this client.
            stream.BeginRead(buffer, 0, Me.BufferSize, AddressOf Read, New ReadAsyncState With {.Client = client, .Buffer = buffer})

            'Notify any listeners that a connection has been made.
            Me.OnConnectionAccepted(New ConnectionEventArgs(host))
        Catch ex As ObjectDisposedException
            'The callback specified when BeginAcceptTcpClient was called gets invoked one last time when the TcpListener is stopped.
            'This exception is thrown when EndAcceptTcpClient is called on a disposed server.
        End Try
    End Sub

    ''' <summary>
    ''' Creates the server and starts listening for incoming connection requests.
    ''' </summary>
    ''' <param name="port">
    ''' The port to listen on.
    ''' </param>
    Private Sub Initialise(ByVal ipAddress As IPAddress, ByVal port As Integer)
        'Listen on the first IPv4 address assigned to the local machine.
        Me.server = New TcpListener(ipAddress, port)
        Me.server.Start()

        'Get the port number from the server in case a random port was used.
        Me._port = DirectCast(Me.server.LocalEndpoint, IPEndPoint).Port

        'Start listen asynchronously.
        Me.server.BeginAcceptTcpClient(AddressOf AcceptTcpClient, Nothing)
    End Sub

    ''' <summary>
    ''' Converts an address string to an <see cref="IPAddress" /> instance.
    ''' </summary>
    ''' <param name="address">
    ''' The address to parse.
    ''' </param>
    ''' <returns>
    ''' An <b>IPAddress</b> corresponding to the specified address.
    ''' </returns>
    ''' <remarks>
    ''' If <i>address</i> is null or empty or is equal to the machine name then <see cref="IPAddress.Any">Any</see> is returned.
    ''' If <i>address</i> is equal to "localhost" then <see cref="IPAddress.Loopback">Loopback</see> is returned.
    ''' In each case comparisons are case-insensitive.
    ''' </remarks>
    Private Function ParseAddress(ByVal address As String) As IPAddress
        Dim result As IPAddress

        If String.IsNullOrEmpty(address) OrElse _
           address.Equals(Environment.MachineName, StringComparison.CurrentCultureIgnoreCase) Then
            result = IPAddress.Any
        ElseIf address.Equals("localhost", StringComparison.CurrentCultureIgnoreCase) Then
            result = IPAddress.Loopback
        Else
            result = IPAddress.Parse(address)
        End If

        Return result
    End Function

    ''' <summary>
    ''' Receives an incoming message.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the read operation.
    ''' </param>
    Private Sub Read(ByVal ar As IAsyncResult)
        Dim asyncState = DirectCast(ar.AsyncState, ReadAsyncState)
        Dim buffer = asyncState.Buffer
        Dim client = asyncState.Client

        Try
            Dim stream = client.GetStream()

            'Complete the asynchronous read and get the first block of data.
            Dim byteCount = stream.EndRead(ar)

            If byteCount = 0 Then
                'If there is no data when an asynchronous read completes it is because the client closed the connection.
                Me.RemoveClient(client)
            Else
                'Start building the message.
                Dim message As New StringBuilder(Me.Encoding.GetString(buffer, 0, byteCount))

                'As long as there is more data...
                While stream.DataAvailable
                    '...read another block of data.
                    byteCount = stream.Read(buffer, 0, Me.BufferSize)

                    'Build the message block by block.
                    message.Append(Me.Encoding.GetString(buffer, 0, byteCount))
                End While

                'Listen asynchronously for another incoming message.
                stream.BeginRead(buffer, 0, Me.BufferSize, AddressOf Read, New ReadAsyncState With {.Client = client, .Buffer = buffer})

                'Notify any listeners that a message was received.
                Me.OnMessageReceived(New MessageReceivedEventArgs(Me.clients(client), message.ToString()))
            End If
        Catch ex As InvalidOperationException
            'The callback specified when BeginRead was called gets invoked one last time when the TcpListener is stopped.
            'This exception is thrown when GetStream is called on a disconnected client or EndRead is called on a disposed stream.
        End Try
    End Sub

    ''' <summary>
    ''' Removes a client from the list of connected clients.
    ''' </summary>
    ''' <param name="client">
    ''' The client to remove.
    ''' </param>
    Private Sub RemoveClient(ByVal client As TcpClient)
        Dim host = Me.clients(client)

        Me.clients.Remove(client)

        'Notify any listeners that the host has disconnected.
        Me.OnConnectionClosed(New ConnectionEventArgs(host))
    End Sub

    ''' <summary>
    ''' Sends a message to a client.
    ''' </summary>
    ''' <param name="client">
    ''' The client to send the message to.
    ''' </param>
    ''' <param name="message">
    ''' The message to send.
    ''' </param>
    Private Sub Send(ByVal client As TcpClient, ByVal message As String)
        Dim stream = client.GetStream()
        Dim buffer As Byte() = Me.Encoding.GetBytes(message)

        'Send the message asynchronously.
        stream.BeginWrite(buffer, 0, buffer.Length, AddressOf Write, stream)
    End Sub

    ''' <summary>
    ''' Completes writing a message to a client.
    ''' </summary>
    ''' <param name="ar">
    ''' Contains state information for the write operation.
    ''' </param>
    Private Sub Write(ByVal ar As IAsyncResult)
        'Complete the asynchronous write.
        DirectCast(ar.AsyncState, NetworkStream).EndWrite(ar)
    End Sub

#End Region 'Private Methods

#End Region 'Methods

End Class
